Attribute VB_Name = "modGeometria"
'==================================================================
' Descrizione.....: Libreria di routines geometriche.
' Nome dei Files..: modGeometria.bas
' Data............: 19/9/2001
' Versione........: 3.0 a 32 bits.
' Sistema.........: VB6 (SP5) sotto Windows NT (SP6).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.virgilio.it/flanguasco/
'==================================================================
'
'   Traduzione dalle routines FORTRAN di: John Burkardt.
'
Option Explicit
Public Sub polygon_centroid_3d(ByVal N#, X#(), Y#(), Z#(), _
                               ByRef cx#, ByRef cy#, ByRef cz#)
'
'*******************************************************************************
'
' POLYGON_CENTROID_3D computes the centroid of a polygon in 3D.
'
'
'  Method:
'
'    The centroid is the area-weighted sum of the centroids of
'    disjoint triangles that make up the polygon.
'
'  Reference:
'
'    Adrian Bowyer and John Woodwark,
'    A Programmer's Geometry,
'    Butterworths, 1983.
'
'  Modified:
'
'    16 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, integer N, the number of vertices of the polygon.
'
'    Input, real X(N), Y(N), Z(N), the coordinates of the vertices.
'
'    Output, real CX, CY, CZ, the coordinates of the centroid.
'
    Dim I&, area#, areat#
'
    area = 0#
    cx = 0#
    cy = 0#
    cz = 0#

    For I = 1 To N - 2

        Call triangle_area_3d(X(I), Y(I), Z(I), X(I + 1), _
                              Y(I + 1), Z(I + 1), X(N), Y(N), Z(N), areat)
'
        area = area + areat
        cx = cx + areat * (X(I) + X(I + 1) + X(N)) / 3#
        cy = cy + areat * (Y(I) + Y(I + 1) + Y(N)) / 3#
        cz = cz + areat * (Z(I) + Z(I + 1) + Z(N)) / 3#
    Next I
'
    cx = cx / area
    cy = cy / area
    cz = cz / area
'
'
'
End Sub
Public Sub triangle_area_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                            ByVal X2#, ByVal Y2#, ByVal Z2#, _
                            ByVal X3#, ByVal Y3#, ByVal Z3#, ByRef area#)
'
'*******************************************************************************
'
' TRIANGLE_AREA_3D computes the area of a triangle in 3D.
'
'
'  Discussion:
'
'    This routine uses the fact that the norm of the cross product vector
'    is the area of the parallelogram they form.  The triangle they
'    form has half that area.
'
'  Reference:
'
'    Adrian Bowyer and John Woodwark,
'    A Programmer's Geometry,
'    Butterworths, 1983.
'
'  Modified:
'
'    06 August 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, the triangle vertices.
'
'    Output, real AREA, the area of the triangle.
'
    Dim X4#, Y4#, Z4#, norm#
'
    Call cross_3d(X2 - X1, Y2 - Y1, Z2 - Z1, X3 - X1, Y3 - Y1, Z3 - Z1, X4, Y4, Z4)
'
    norm = enorm_3d(X4, Y4, Z4)
'
    area = 0.5 * norm
'
'
'
End Sub
Public Sub cross_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                    ByVal X2#, ByVal Y2#, ByVal Z2#, _
                    ByRef X3#, ByRef Y3#, ByRef Z3#)
'
'*******************************************************************************
'
' CROSS_3D computes the cross product of two vectors in 3D.
'
'
'  Definition:
'
'    The cross product in 3D can be regarded as the determinant of the
'    symbolic matrix:
'
'          |  i  j  k |
'      det | x1 y1 z1 |
'          | x2 y2 z2 |
'
'      = ( y1 * z2 - z1 * y2 ) * i
'      + ( z1 * x2 - x1 * z2 ) * j
'      + ( x1 * y2 - y1 * x2 ) * k
'
'  Modified:
'
'    19 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, the coordinates of the vectors.
'
'    Output, real X3, Y3, Z3, the cross product vector.
'
    X3 = Y1 * Z2 - Z1 * Y2
    Y3 = Z1 * X2 - X1 * Z2
    Z3 = X1 * Y2 - Y1 * X2
'
'
'
End Sub
Public Function enorm_3d(ByVal X1#, ByVal Y1#, ByVal Z1#) As Double
'
'*******************************************************************************
'
' ENORM_3D computes the Euclidean norm of a vector in 3D.
'
'
'  Modified:
'
'    27 July 1998
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, the coordinates of the vector.
'
'    Output, real ENORM_3D, the Euclidean norm of the vector.
'
    enorm_3d = Sqr(X1 * X1 + Y1 * Y1 + Z1 * Z1)
'
'
'
End Function
Public Sub line_seg_contains_point_3D(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                      ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                      ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                      ByRef u#, ByRef v#)
'
'*******************************************************************************
'
' LINE_SEG_CONTAINS_POINT_3D reports if a line segment contains a point in 3D.
'
'   Nota: questa routine e' derivata dalla line_seg_contains_point_2d
'         e NON e' presente nella libreria originale.
'
'  Discussion:
'
'    In exact arithmetic, point P3 = (X3,Y3,Z3) is on the line segment between
'    P1=(X1,Y1,Z1) and P2=(X2,Y2,Z2) if and only if 0 <= U <= 1 and V = 0.
'
'  Modified:
'
'    28 September 2000
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2 the endpoints P1 and P2 of a line segment.
'
'    Input, real X3, Y3, Z3 a point P3 to be tested.
'
'    Output, real U, the coordinate of (X3,Y3,Z3) along the axis from
'    with origin at P1 and unit at P2.
'
'    Output, real V, the magnitude of the off-axis portion of the
'    vector P3-P1, measured in units of (P2-P1).
'
    Dim unit#
'
    unit = Sqr((X2 - X1) ^ 2 + (Y2 - Y1) ^ 2 + (Z2 - Z1) ^ 2)
'
    If (unit = 0#) Then
        If (X3 = X1 And Y3 = Y1 And Z3 = Z1) Then
            u = 0.5
            v = 0#
        Else
            u = 0.5
            v = 1E+38 'huge(v)
        End If
'
    Else
    u = ((X3 - X1) * (X2 - X1) + (Y3 - Y1) * (Y2 - Y1) + (Z3 - Z1) * (Z2 - Z1)) / unit ^ 2
    v = Sqr(((u - 1#) * X1 - u * X2 + X3) ^ 2 _
          + ((u - 1#) * Y1 - u * Y2 + Y3) ^ 2 _
          + ((u - 1#) * Z1 - u * Z2 + Z3) ^ 2 _
          ) / unit
    End If
'
'
'
End Sub
Public Sub line_seg_point_dist_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                  ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                  ByVal X#, ByVal Y#, ByVal Z#, _
                                  ByRef dist#)
'
'*******************************************************************************
'
' LINE_SEG_POINT_DIST_3D: distance ( line segment, point ) in 3D.
'
'
'  Modified:
'
'    02 November 1998
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, the endpoints of the line
'    segment.
'
'    Input, real X, Y, Z, the point whose nearest neighbor on the line
'    segment is to be determined.
'
'    Output, real DIST, the distance from the point to the line segment.
'
  Dim bot#, T#, xn#, yn#, zn#
'
'  If the line segment is actually a point, then the answer is easy.

    If (X1 = X2 And Y1 = Y2 And Z1 = Z2) Then
        xn = X1
        yn = Y1
        zn = Z1
    Else
'
        bot = enormsq0_3d(X1, Y1, Z1, X2, Y2, Z2)
'
        T = ((X1 - X) * (X1 - X2) _
           + (Y1 - Y) * (Y1 - Y2) _
           + (Z1 - Z) * (Z1 - Z2)) / bot
'
        T = DMAX1(T, 0#)
        T = DMIN1(T, 1#)
'
        xn = X1 + T * (X2 - X1)
        yn = Y1 + T * (Y2 - Y1)
        zn = Z1 + T * (Z2 - Z1)
    End If
'
    dist = enorm0_3d(X, Y, Z, xn, yn, zn)
'
'
'
End Sub
Public Function enorm0_3d(ByVal X0, ByVal Y0, ByVal Z0, _
                          ByVal X1, ByVal Y1, ByVal Z1) As Double
'
'*******************************************************************************
'
' ENORM0_3D computes the Euclidean norm of (P1-P0) in 3D.
'
'
'  Modified:
'
'    16 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X0, Y0, Z0, X1, Y1, Z1, the coordinates of the points
'    P0 and P1.
'
'    Output, real ENORM0_3D, the Euclidean norm of (P1-P0).
'
    enorm0_3d = Sqr((X1 - X0) ^ 2 + (Y1 - Y0) ^ 2 + (Z1 - Z0) ^ 2)
'
'
'
End Function
Private Function DMAX1(ParamArray Vd() As Variant) As Double
'
'   Implementa la funzione DMAX1(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(Vd)
    J2 = UBound(Vd)
    vDMax = Vd(J1)
    For J = J1 + 1 To J2
        If vDMax < Vd(J) Then vDMax = Vd(J)
    Next J
'
    DMAX1 = CDbl(vDMax)
'
'
'
End Function
Private Function DMIN1(ParamArray Vd() As Variant) As Double
'
'   Implementa la funzione DMIN1(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(Vd)
    J2 = UBound(Vd)
    vDMin = Vd(J1)
    For J = J1 + 1 To J2
        If vDMin > Vd(J) Then vDMin = Vd(J)
    Next J
'
    DMIN1 = CDbl(vDMin)
'
'
'
End Function
Private Function enormsq0_3d(ByVal X0, ByVal Y0, ByVal Z0, _
                             ByVal X1, ByVal Y1, ByVal Z1) As Double
'
'*******************************************************************************
'
' ENORMSQ0_3D computes the square of the Euclidean norm of (P1-P0) in 3D.
'
'
'  Modified:
'
'    06 February 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X0, Y0, Z0, X1, Y1, Z1, the coordinates of the points
'    P0 and P1.
'
'    Output, real ENORMSQ0_3D, the square of the Euclidean norm of (P1-P0).
'
    enormsq0_3d = (X1 - X0) ^ 2 + (Y1 - Y0) ^ 2 + (Z1 - Z0) ^ 2
'
'
'
End Function
Public Sub line_seg_contains_point_2d(ByVal X1#, ByVal Y1#, _
                                      ByVal X2#, ByVal Y2#, _
                                      ByVal X3#, ByVal Y3#, _
                                      ByRef u#, ByRef v#)
'
'*******************************************************************************
'
' LINE_SEG_CONTAINS_POINT_2D reports if a line segment contains a point in 2D.
'
'
'  Discussion:
'
'    In exact arithmetic, point P3 = (X3,Y3) is on the line segment between
'    P1=(X1,Y1) and P2=(X2,Y2) if and only if 0 <= U <= 1 and V = 0.
'
'  Modified:
'
'    28 September 2000
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, X2, Y2, the endpoints P1 and P2 of a line segment.
'
'    Input, real X3, Y3, a point P3 to be tested.
'
'    Output, real U, the coordinate of (X3,Y3) along the axis from
'    with origin at P1 and unit at P2.
'
'    Output, real V, the magnitude of the off-axis portion of the
'    vector P3-P1, measured in units of (P2-P1).
'
    Dim unit#
'
    unit = Sqr((X2 - X1) ^ 2 + (Y2 - Y1) ^ 2)

    If (unit = 0#) Then
        If (X3 = X1 And Y3 = Y1) Then
            u = 0.5
            v = 0#
        Else
            u = 0.5
            v = 1E+38 'huge(v)
        End If
'
    Else
    u = ((X3 - X1) * (X2 - X1) + (Y3 - Y1) * (Y2 - Y1)) / unit ^ 2
    v = Sqr(((u - 1#) * X1 - u * X2 + X3) ^ 2 _
             + ((u - 1#) * Y1 - u * Y2 + Y3) ^ 2) / unit
    End If
'
'
'
End Sub
Public Sub plane_exp2imp_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                            ByVal X2#, ByVal Y2#, ByVal Z2#, _
                            ByVal X3#, ByVal Y3#, ByVal Z3#, _
                            ByRef a#, ByRef b#, ByRef c#, ByRef d#)
'
'*******************************************************************************
'
' PLANE_EXP2IMP_3D converts an explicit plane to implicit form in 3D.
'
'
'  Definition:
'
'    The explicit form of a plane in 3D is
'
'      (X1,Y1,Z1), (X2,Y2,Z2), (X3,Y3,Z3).
'
'    The implicit form of a plane in 3D is
'
'      A * X + B * Y + C * Z + D = 0
'
'  Reference:
'
'    Adrian Bowyer and John Woodwark,
'    A Programmer's Geometry,
'    Butterworths, 1983.
'
'  Modified:
'
'    04 February 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, X2, X3, Y3, Z3, are three points
'    on the plane, which must be distinct, and not collinear.
'
'    Output, real A, B, C, D, coefficients which describe the plane.
'
    a = (Y2 - Y1) * (Z3 - Z1) - (Z2 - Z1) * (Y3 - Y1)
    b = (Z2 - Z1) * (X3 - X1) - (X2 - X1) * (Z3 - Z1)
    c = (X2 - X1) * (Y3 - Y1) - (Y2 - Y1) * (X3 - X1)
    d = -X2 * a - Y2 * b - Z2 * c
'
'
'
End Sub
Public Function plane_exp2norm_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                  ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                  ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                  ByRef xp#, ByRef yp#, ByRef zp#, _
                                  ByRef xn#, ByRef yn#, ByRef zn#) As Boolean
'
'*******************************************************************************
'
' PLANE_EXP2NORM_3D converts an explicit plane to normal form in 3D.
'
'
'  Definition:
'
'    The explicit form of a plane in 3D is
'
'      (X1,Y1,Z1), (X2,Y2,Z2), (X3,Y3,Z3).
'
'    The normal form of a plane in 3D is
'
'      (Xp,Yp,Zp), a point on the plane, and
'      (Xn,Yn,Zn), the unit normal to the plane.
'
'  Modified:
'
'    02 June 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, X2, X3, Y3, Z3, are three points
'    on the plane, which must be distinct, and not collinear.
'
'    Output, real XP, YP, ZP, a point on the plane.
'
'    Output, real XN, YN, ZN, the unit normal vector to the plane.
'
    Dim norm#
'
    xp = X1
    yp = Y1
    zp = Z1
'
    xn = (Y2 - Y1) * (Z3 - Z1) - (Z2 - Z1) * (Y3 - Y1)
    yn = (Z2 - Z1) * (X3 - X1) - (X2 - X1) * (Z3 - Z1)
    zn = (X2 - X1) * (Y3 - Y1) - (Y2 - Y1) * (X3 - X1)
'
    norm = Sqr(xn * xn + yn * yn + zn * zn)
'
    If (norm = 0#) Then
        'write ( *, * ) ' '
        'write ( *, * ) 'PLANE_EXP2NORM_3D - Fatal error''
        'write ( *, * ) '  The normal vector is null.'
        'write ( *, * ) '  Two points coincide, or nearly so.'
        plane_exp2norm_3d = False
        Exit Function
    End If
'
    xn = xn / norm
    yn = yn / norm
    zn = zn / norm
'
    plane_exp2norm_3d = True
'
'
'
End Function
Public Function plane_exp_normal_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                    ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                    ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                    ByRef xn#, ByRef yn#, zn#) As Boolean
'
'*******************************************************************************
'
' PLANE_EXP_NORMAL_3D finds the normal to an explicit plane in 3D.
'
'
'  Definition:
'
'    The explicit form of a plane in 3D is
'
'      (X1,Y1,Z1), (X2,Y2,Z2), (X3,Y3,Z3).
'
'  Modified:
'
'    16 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, the coordinates
'    of three points that constitute a line.  These points should not
'    be identical, nor colinear.
'
'    Output, real XN, YN, ZN, the coordinates of the unit normal
'    vector to the plane containing the three points.
'
    Dim temp#
'
'  The cross product (P2-P1) x (P3-P1) is a vector normal to
'  (P2-P1) and (P3-P1).
'
    Call cross0_3d(X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, xn, yn, zn)
'
    temp = Sqr(xn * xn + yn * yn + zn * zn)
'
    If (temp = 0#) Then
        'write ( *, * ) ' '
        'write ( *, * ) 'PLANE_EXP_NORMAL_3D - Fatal error''
        'write ( *, * ) '  The plane is poorly defined.'
        plane_exp_normal_3d = False
        Exit Function
    Else
        xn = xn / temp
        yn = yn / temp
        zn = zn / temp
    End If
'
    plane_exp_normal_3d = True
'
'
'
End Function
Public Sub cross0_3d(ByVal X0#, ByVal Y0#, ByVal Z0#, _
                     ByVal X1#, ByVal Y1#, ByVal Z1#, _
                     ByVal X2#, ByVal Y2#, ByVal Z2#, _
                     ByRef X3#, ByRef Y3#, ByRef Z3#)
'
'*******************************************************************************
'
' CROSS0_3D computes the cross product of (P1-P0) and (P2-P0) in 3D.
'
'
'  Discussion:
'
'    The vectors are specified with respect to a basis point P0.
'    We are computing the normal to the triangle (P0,P1,P2).
'
'  Modified:
'
'    19 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X0, Y0, Z0, X1, Y1, Z1, X2, Y2, Z2, the coordinates of
'    three points.  The basis point is (X0,Y0,Z0).
'
'    Output, real X3, Y3, Z3, the cross product (P1-P0) x (P2-P0), or
'    (X1-X0,Y1-Y0,Z1-Z0) x (X2-X0,Y2-Y0,Z2-Z0).
'
    X3 = (Y1 - Y0) * (Z2 - Z0) - (Z1 - Z0) * (Y2 - Y0)
    Y3 = (Z1 - Z0) * (X2 - X0) - (X1 - X0) * (Z2 - Z0)
    Z3 = (X1 - X0) * (Y2 - Y0) - (Y1 - Y0) * (X2 - X0)
'
'
'
End Sub
Public Sub plane_exp_point_dist_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                   ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                   ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                   ByVal X#, ByVal Y#, ByVal Z#, ByRef dist#)
'
'*******************************************************************************
'
' PLANE_EXP_POINT_DIST_3D: distance ( explicit plane, point ) in 3D.
'
'
'  Definition:
'
'    The explicit form of a plane in 3D is
'
'      (X1,Y1,Z1), (X2,Y2,Z2), (X3,Y3,Z3).
'
'  Modified:
'
'    15 July 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, are the
'    coordinates of three points on the plane.
'
'    Input, real X, Y, Z, the coordinates of the point.
'
'    Output, real DIST, the distance from the point to the plane.
'
    Dim a#, b#, c#, d#
'
    Call plane_exp2imp_3d(X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, a, b, c, d)
'
    Call plane_imp_point_dist_3d(a, b, c, d, X, Y, Z, dist)
'
'
'
End Sub
Public Function plane_imp_point_dist_3d(ByVal a#, ByVal b#, ByVal c#, ByVal d#, _
                                        ByVal X#, ByVal Y#, ByVal Z#, _
                                        ByRef dist#) As Boolean
'
'*******************************************************************************
'
' PLANE_IMP_POINT_DIST_3D: distance ( implicit plane, point ) in 3D.
'
'
'  Definition:
'
'    The implicit form of a plane in 3D is:
'
'      A * X + B * Y + C * Z + D = 0
'
'  Reference:
'
'    Adrian Bowyer and John Woodwark,
'    A Programmer's Geometry,
'    Butterworths, 1983.
'
'  Modified:
'
'    23 June 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real A, B, C, D, coefficients that define the plane as
'    the set of points for which A*X+B*Y+C*Z+D = 0.
'
'    Input, real X, Y, Z, the coordinates of the point.
'
'    Output, real DIST, the distance from the point to the plane.
'
    Dim norm#
'
    norm = Sqr(a * a + b * b + c * c)
'
    If (norm = 0#) Then
        'write ( *, * ) ' '
        'write ( *, * ) 'PLANE_IMP_POINT_DIST_3D - Fatal error''
        'write ( *, * ) '  The plane normal vector is null.'
        plane_imp_point_dist_3d = False
        Exit Function
    End If
'
    dist = Abs(a * X + b * Y + c * Z + d) / norm
'
    plane_imp_point_dist_3d = True
'
'
'
End Function
Public Sub plane_imp_triangle_int_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                     ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                     ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                     ByVal a#, ByVal b#, ByVal c#, ByVal d#, _
                                     ByRef num_int&, X#(), Y#(), Z#())
'
'*******************************************************************************
'
' PLANE_IMP_TRIANGLE_INT_3D: intersection ( implicit plane, triangle ) in 3D.
'
'
'  Definition:
'
'    The implicit form of a plane in 3D is:
'
'      A * X + B * Y + C * Z + D = 0
'
'  Discussion:
'
'    There may be 0, 1, 2 or 3 points of intersection returned.
'
'    If two intersection points are returned, then the entire line
'    between them comprises points of intersection.
'
'    If three intersection points are returned, then all points of
'    the triangle intersect the plane.
'
'  Modified:
'
'    02 June 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, the coordinates
'    of the vertices of the triangle.
'
'    Input, real A, B, C, D, the parameters that define the implicit plane.
'
'    Output, integer NUM_INT, the number of intersection points returned.
'
'    Output, real X(3), Y(3), Z(3), the coordinates of the intersection points.
'
    Dim dist1#, dist2#, dist3#
'
    num_int = 0
'
'  Compute the signed distances between the vertices and the plane.
'
    dist1 = a * X1 + b * Y1 + c * Z1 + d
    dist2 = a * X2 + b * Y2 + c * Z2 + d
    dist3 = a * X3 + b * Y3 + c * Z3 + d
'
'  Consider any zero distances.
'
    If (dist1 = 0#) Then
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = X1
        Y(num_int) = Y1
        Z(num_int) = Z1
    End If
'
    If (dist2 = 0#) Then
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = X2
        Y(num_int) = Y2
        Z(num_int) = Z2
    End If
'
    If (dist3 = 0#) Then
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = X3
        Y(num_int) = Y3
        Z(num_int) = Z3
    End If
'
'  If 2 or 3 of the nodes intersect, we're already done.
'
    If (num_int >= 2) Then
        Exit Sub
    End If
'
'  If one node intersects, then we're done unless the other two
'  are of opposite signs.
'
    If (num_int = 1) Then
        If (dist1 = 0#) Then
            Call plane_triangle_int_add_3d(X2, Y2, Z2, X3, Y3, Z3, _
                                           dist2, dist3, num_int, X(), Y(), Z())
'
        ElseIf (dist2 = 0#) Then
            Call plane_triangle_int_add_3d(X1, Y1, Z1, X3, Y3, Z3, _
                                           dist1, dist3, num_int, X(), Y(), Z())
'
        ElseIf (dist3 = 0#) Then
            Call plane_triangle_int_add_3d(X1, Y1, Z1, X2, Y2, Z2, _
                                           dist1, dist2, num_int, X(), Y(), Z())
        End If
        Exit Sub
    End If
'
'  All nodal distances are nonzero, and there is at least one
'  positive and one negative.
'
    If (dist1 * dist2 < 0# And dist1 * dist3 < 0#) Then
        Call plane_triangle_int_add_3d(X1, Y1, Z1, X2, Y2, Z2, _
                                       dist1, dist2, num_int, X(), Y(), Z())
        Call plane_triangle_int_add_3d(X1, Y1, Z1, X3, Y3, Z3, _
                                       dist1, dist3, num_int, X(), Y(), Z())
'
    ElseIf (dist2 * dist1 < 0# And dist2 * dist3 < 0#) Then
        Call plane_triangle_int_add_3d(X2, Y2, Z2, X1, Y1, Z1, _
                                       dist2, dist1, num_int, X(), Y(), Z())
        Call plane_triangle_int_add_3d(X2, Y2, Z2, X3, Y3, Z3, _
                                       dist2, dist3, num_int, X(), Y(), Z())
'
    ElseIf (dist3 * dist1 < 0# And dist3 * dist2 < 0#) Then
        Call plane_triangle_int_add_3d(X3, Y3, Z3, X1, Y1, Z1, _
                                       dist3, dist1, num_int, X(), Y(), Z())
        Call plane_triangle_int_add_3d(X3, Y3, Z3, X2, Y2, Z2, _
                                       dist3, dist2, num_int, X(), Y(), Z())
    End If
'
'
'
End Sub
Public Sub plane_imp_triangle_near_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                      ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                      ByVal X3#, ByVal Y3#, ByVal Z3#, _
                                      ByVal a#, ByVal b#, ByVal c#, ByVal d#, _
                                      ByRef dist#, ByRef num_near&, _
                                      ByRef X#(), ByRef Y#(), ByRef Z#())
'
'*******************************************************************************
'
' PLANE_IMP_TRIANGLE_NEAR_3D: nearest ( implicit plane, triangle ) in 3D.
'
'
'  Definition:
'
'    The implicit form of a plane in 3D is:
'
'      A * X + B * Y + C * Z + D = 0
'
'  Comments:
'
'    Please see to it that the underlying distance routine always returns
'    one of the endpoints if the entire line segment is at zero distance.
'
'  Modified:
'
'    18 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, the coordinates
'    of the vertices of the triangle.
'
'    Input, real A, B, C, D, the parameters that define the implicit plane.
'
'    Output, real DIST, the distance between the triangle and the plane.
'
'    Output, integer NUM_NEAR, the number of nearest points returned.
'
'    Output, real X(6), Y(6), Z(6), a collection of nearest points.
'
'    If DIST = 0, then each point is a point of intersection, and there
'    will be at most 3 such points returned.
'
'    If DIST > 0, then the points are listed in pairs, with the first
'    being on the triangle, and the second on the plane.  Two points will
'    be listed in the most common case, but possibly 4 or 6.
'
    Dim dist12#, dist23#, dist31#
    Dim xp#, xt#
    Dim yp#, yt#
    Dim zp#, zt#
'
    num_near = 0
'
'  Consider the line segment P1 - P2.
'
    Call plane_imp_line_seg_near_3d(X1, Y1, Z1, X2, Y2, Z2, _
                                    a, b, c, d, dist12, xp, yp, zp, xt, yt, zt)
'
    dist = dist12
'
    num_near = num_near + 1
    X(num_near) = xt
    Y(num_near) = yt
    Z(num_near) = zt
'
    If (dist12 > 0#) Then
        num_near = num_near + 1
        X(num_near) = xp
        Y(num_near) = yp
        Z(num_near) = zp
    End If
'
'  Consider the line segment P2 - P3.
'
    Call plane_imp_line_seg_near_3d(X2, Y2, Z2, X3, Y3, Z3, _
                                    a, b, c, d, dist23, xp, yp, zp, xt, yt, zt)
'
    If (dist23 < dist) Then
        num_near = 0
        dist = dist23
'
        num_near = num_near + 1
        X(num_near) = xt
        Y(num_near) = yt
        Z(num_near) = zt
'
        If (dist23 > 0#) Then
            num_near = num_near + 1
            X(num_near) = xp
            Y(num_near) = yp
            Z(num_near) = zp
        End If
'
    ElseIf (dist23 = dist) Then
        num_near = num_near + 1
        X(num_near) = xt
        Y(num_near) = yt
        Z(num_near) = zt
'
        If (dist23 > 0#) Then
            num_near = num_near + 1
            X(num_near) = xp
            Y(num_near) = yp
            Z(num_near) = zp
        End If
    End If
'
'  Consider the line segment P3 - P1.
'
    Call plane_imp_line_seg_near_3d(X3, Y3, Z3, X1, Y1, Z1, _
                                    a, b, c, d, dist31, xp, yp, zp, xt, yt, zt)
    If (dist31 < dist) Then
        num_near = 0
        dist = dist31
'
        num_near = num_near + 1
        X(num_near) = xt
        Y(num_near) = yt
        Z(num_near) = zt
'
        If (dist31 > 0#) Then
            num_near = num_near + 1
            X(num_near) = xp
            Y(num_near) = yp
            Z(num_near) = zp
        End If
'
    ElseIf (dist31 = dist) Then
        num_near = num_near + 1
        X(num_near) = xt
        Y(num_near) = yt
        Z(num_near) = zt
'
        If (dist31 > 0#) Then
            num_near = num_near + 1
            X(num_near) = xp
            Y(num_near) = yp
            Z(num_near) = zp
        End If
  End If
'
'
'
End Sub
Public Function plane_imp_line_seg_near_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                           ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                           ByVal a#, ByVal b#, ByVal c#, ByVal d#, _
                                           ByRef dist#, ByRef xp#, ByRef yp#, ByRef zp#, _
                                           ByRef xls#, ByRef yls#, ByRef zls#) As Boolean
'
'*******************************************************************************
'
' PLANE_IMP_LINE_SEG_NEAR_3D: nearest ( implicit plane, line segment ) in 3D.
'
'
'  Definition:
'
'    The implicit form of a plane in 3D is:
'
'      A * X + B * Y + C * Z + D = 0
'
'  Modified:
'
'    17 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, the endpoints of the line
'    segment.
'
'    Input, real A, B, C, D, the parameters that define the implicit
'    plane.
'
'    Output, real DIST, the distance between the line segment and
'    the plane.
'
'    Output, real XP, YP, ZP, the nearest point on the plane.
'
'    Output, real XLS, YLS, ZLS, the nearest point on the line segment
'    to the plane.  If DIST is zero, the (XLS,YLS,ZLS) is a point of
'    intersection of the plane and the line segment.
'
    Dim Alpha#, an#, bn#, cn#, dn#, norm#
    Dim p1#, p2#
'
    xls = 0#
    yls = 0#
    zls = 0#
    xp = 0#
    yp = 0#
    zp = 0#
'
    norm = Sqr(a * a + b * b + c * c)
    
    If (norm = 0#) Then
        'write ( *, * ) ' '
        'write ( *, * ) 'PLANE_IMP_LINE_SEG_NEAR_3D - Fatal error''
        'write ( *, * ) '  Plane normal vector is null.'
        plane_imp_line_seg_near_3d = False
        Exit Function
    End If
'
'  The normalized coefficients allow us to compute the (signed) distance.
'
    an = a / norm
    bn = b / norm
    cn = c / norm
    dn = d / norm
'
'  If the line segment is actually a point, then the answer is easy.
'
    If (X1 = X2 And Y1 = Y2 And Z1 = Z2) Then
        p1 = an * X1 + bn * Y1 + cn * Z1 + dn
        dist = Abs(p1)
        xls = X1
        yls = Y1
        zls = Z1
        xp = xls - an * p1
        yp = yls - bn * p1
        zp = zls - cn * p1
        plane_imp_line_seg_near_3d = True
        Exit Function
  End If
'
'  Compute the projections of the two points onto the normal vector.
'
    p1 = an * X1 + bn * Y1 + cn * Z1 + dn
    p2 = an * X2 + bn * Y2 + cn * Z2 + dn
'
'  If these have the same sign, then the line segment does not
'  cross the plane, and one endpoint is the nearest point.
'
    If ((p1 > 0# And p2 > 0#) Or _
        (p1 < 0# And p2 < 0#)) Then
'
        p1 = Abs(p1)
        p2 = Abs(p2)

        If (p1 < p2) Then
            xls = X1
            yls = Y1
            zls = Z1
            xp = xls - an * p1
            yp = yls - bn * p1
            zp = zls - cn * p1
            dist = p1
        Else
            xls = X2
            yls = Y2
            zls = Z2
            dist = p2
            xp = xls - an * p2
            yp = yls - bn * p2
            zp = zls - cn * p2
        End If
'
'  If the projections differ in sign, the line segment crosses the plane.
'
    Else
        If (p1 = 0#) Then
            Alpha = 0#
            ElseIf (p2 = 0#) Then
            Alpha = 1#
        Else
            Alpha = p2 / (p2 - p1)
        End If

        xls = Alpha * X1 + (1# - Alpha) * X2
        yls = Alpha * Y1 + (1# - Alpha) * Y2
        zls = Alpha * Z1 + (1# - Alpha) * Z2
        xp = xls
        yp = yls
        zp = zls

        dist = 0#
    End If
'
    plane_imp_line_seg_near_3d = True
'
'
'
End Function
Private Sub plane_triangle_int_add_3d(ByVal X1#, ByVal Y1#, ByVal Z1#, _
                                      ByVal X2#, ByVal Y2#, ByVal Z2#, _
                                      ByVal dist1#, ByVal dist2#, _
                                      ByRef num_int&, ByRef X#(), ByRef Y#(), ByRef Z#())
'
'*******************************************************************************
'
' PLANE_TRIANGLE_INT_ADD_3D is a utility for plane/triangle intersections.
'
'
'  Discussion:
'
'    This routine is called to consider the value of the signed distance
'    from a plane of two nodes of a triangle.  If the two values
'    have opposite signs, then there is a point of intersection between
'    them.  The routine computes this point and adds it to the list.
'
'  Modified:
'
'    19 April 1999
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input, real X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, the coordinates
'    of two vertices of a triangle.
'
'    Input, real DIST1, DIST2, the signed distances of the two vertices
'    from a plane.
'
'    Input/output, integer NUM_INT, the number of intersection points.
'
'    Input/output, real X(NUM_INT), Y(NUM_INT), Z(NUM_INT), the coordinates
'    of the intersection points.
'
    Dim Alpha#
'
    If (dist1 = 0#) Then
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = X1
        Y(num_int) = Y1
        Z(num_int) = Z1
'
    ElseIf (dist2 = 0#) Then
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = X2
        Y(num_int) = Y2
        Z(num_int) = Z2
'
    ElseIf (dist1 * dist2 < 0#) Then
        Alpha = dist2 / (dist2 - dist1)
        num_int = num_int + 1
        ReDim Preserve X(1 To num_int), Y(1 To num_int), Z(1 To num_int)
        X(num_int) = Alpha * X1 + (1# - Alpha) * X2
        Y(num_int) = Alpha * Y1 + (1# - Alpha) * Y2
        Z(num_int) = Alpha * Z1 + (1# - Alpha) * Z2
    End If
'
'
'
End Sub
